@@ -17,9 +17,6 @@ def outline_to_str(outline: Outline)->str:
 def str_to_outline(outline_str: str)->Outline:
 	return tuple(outline_str.split("/"))
 
-def count_str_outline_len(outline_str: str)->int:
-	return outline_str.count("/")+1
-
 F=TypeVar("F", bound=Callable)
 def with_lock(f: F)->F:
 	@functools.wraps(f)
@@ -99,10 +96,9 @@ class SQLiteDictionaryBase(StenoDictionary):
 
 	@with_lock
 	def update_str(self, data: Iterable[Tuple[str, str]])->None:
-		self._cursor.executemany("replace into dict values (?, ?, ?)", (
-			(outline, translation, count_str_outline_len(outline))
-			for outline, translation in data
-			))
+		self._cursor.executemany(
+				"replace into dict (outline, translation, length) "
+				"values (:0, :1, length(:0)-length(replace(:0, '/', ''))+1)", data)
 		self._longest_key=self._compute_longest_key_unlocked()
 
 	def update(self, *args)->None:
@@ -120,7 +116,7 @@ class SQLiteDictionaryBase(StenoDictionary):
 
 	@with_lock
 	def __setitem__(self, outline: Outline, translation: str)->None:
-		self._cursor.execute("replace into dict values (?, ?, ?)",
+		self._cursor.execute("replace into dict (outline, translation, length) values (?, ?, ?)",
 				(outline_to_str(outline), translation, len(outline)))
 		self._longest_key=self._compute_longest_key_unlocked()
 
@@ -171,8 +167,11 @@ class SQLiteDictionary(SQLiteDictionaryBase):
 		if self.path is None:  # type: ignore
 			path_was_none=True
 			self.path=filename
-		self._connect(filename)
-		self._longest_key=self.compute_longest_key()
+		with self._lock:
+			self._connect_unlocked(filename)
+			self._longest_key=self._compute_longest_key_unlocked()
+			if [*self._cursor.execute('select * from sqlite_master where type="table" and name="readonly"')]:
+				self.readonly=True
 		if path_was_none:
 			self.path=None  # type: ignore
 
